@@ -30,7 +30,7 @@ __all__ = ["GraphBuilder"]
 # -------------------------------
 import itertools
 import logging
-from collections import ChainMap
+from collections import ChainMap, defaultdict
 from contextlib import contextmanager
 from dataclasses import dataclass
 from typing import Any, Collection, Dict, Iterable, Iterator, List, Mapping, Optional, Set, Tuple, Union
@@ -46,6 +46,7 @@ from lsst.daf.butler import (
     DimensionGraph,
     DimensionUniverse,
     NamedKeyDict,
+    NamedValueSet,
     Quantum,
     Registry,
 )
@@ -594,6 +595,7 @@ class _PipelineScaffolding:
                 attr,
                 _DatasetDict.fromDatasetTypes(getattr(datasetTypes, attr), universe=registry.dimensions),
             )
+        self.defaultDatasetQueryConstraints = datasetTypes.queryConstraints
         # Aggregate all dimensions for all non-init, non-prerequisite
         # DatasetTypes.  These are the ones we'll include in the big join
         # query.
@@ -655,6 +657,11 @@ class _PipelineScaffolding:
     per-Quantum when generating the graph (`_DatasetDict`).
     """
 
+    defaultDatasetQueryConstraints: NamedValueSet[DatasetType]
+    """Datasets that should be used as constraints in the initial query,
+    according to tasks (`NamedValueSet`).
+    """
+
     dimensions: DimensionGraph
     """All dimensions used by any regular input, intermediate, or output
     (not prerequisite) dataset; the set of dimension used in the "Big Join
@@ -724,7 +731,10 @@ class _PipelineScaffolding:
         # inputs and outputs.  We limit the query to only dimensions that are
         # associated with the input dataset types, but don't (yet) try to
         # obtain the dataset_ids for those inputs.
-        _LOG.debug("Submitting data ID query and materializing results.")
+        _LOG.debug(
+            "Submitting data ID query over dimensions %s and materializing results.",
+            list(self.dimensions.names),
+        )
         queryArgs: Dict[str, Any] = {
             "dimensions": self.dimensions,
             "where": userQuery,
@@ -732,8 +742,11 @@ class _PipelineScaffolding:
             "bind": bind,
         }
         if datasetQueryConstraint == DatasetQueryConstraintVariant.ALL:
-            _LOG.debug("Constraining graph query using all datasets in pipeline.")
-            queryArgs["datasets"] = list(self.inputs)
+            _LOG.debug(
+                "Constraining graph query using default of %s.",
+                list(self.defaultDatasetQueryConstraints.names),
+            )
+            queryArgs["datasets"] = list(self.defaultDatasetQueryConstraints)
             queryArgs["collections"] = collections
         elif datasetQueryConstraint == DatasetQueryConstraintVariant.OFF:
             _LOG.debug("Not using dataset existence to constrain query.")
@@ -766,7 +779,6 @@ class _PipelineScaffolding:
             # quanta and then connecting them to each other.
             n = -1
             for n, commonDataId in enumerate(commonDataIds):
-                _LOG.debug("Next DataID = %s", commonDataId)
                 # Create DatasetRefs for all DatasetTypes from this result row,
                 # noting that we might have created some already.
                 # We remember both those that already existed and those that we
@@ -783,7 +795,6 @@ class _PipelineScaffolding:
                     ref = refs.get(datasetDataId)
                     if ref is None:
                         ref = DatasetRef(datasetType, datasetDataId)
-                        _LOG.debug("Made new ref = %s", ref)
                         refs[datasetDataId] = ref
                     refsForRow[datasetType.name] = ref
                 # Create _QuantumScaffolding objects for all tasks from this
@@ -924,6 +935,16 @@ class _PipelineScaffolding:
 
         resolvedRefQueryResults: Iterable[DatasetRef]
 
+        # Updating constrainedByAllDatasets here is not ideal, but we have a
+        # few different code paths that each transfer different pieces of
+        # information about what dataset query constraints were applied here,
+        # and none of them has the complete picture until we get here.  We're
+        # long overdue for a QG generation rewrite that will make this go away
+        # entirely anyway.
+        constrainedByAllDatasets = (
+            constrainedByAllDatasets and self.defaultDatasetQueryConstraints == self.inputs.keys()
+        )
+
         # Look up [init] intermediate and output datasets in the output
         # collection, if there is an output collection.
         if run_exists or skip_collections_wildcard is not None:
@@ -1034,6 +1055,11 @@ class _PipelineScaffolding:
                         "or the input collections have been modified since "
                         "QuantumGraph generation began."
                     )
+                elif not datasetType.dimensions:
+                    raise RuntimeError(
+                        f"Dataset {datasetType.name!r} (with no dimensions) could not be found in "
+                        f"collections {collections}."
+                    )
                 else:
                     # if the common dataIds were not constrained using all the
                     # input dataset types, it is possible that some data ids
@@ -1210,13 +1236,18 @@ class _PipelineScaffolding:
                     refDict.update(idMaker.resolveDict(refDict))
 
     def makeQuantumGraph(
-        self, metadata: Optional[Mapping[str, Any]] = None, datastore: Optional[Datastore] = None
+        self,
+        registry: Registry,
+        metadata: Optional[Mapping[str, Any]] = None,
+        datastore: Optional[Datastore] = None,
     ) -> QuantumGraph:
         """Create a `QuantumGraph` from the quanta already present in
         the scaffolding data structure.
 
         Parameters
         ---------
+        registry : `lsst.daf.butler.Registry`
+            Registry for the data repository; used for all data ID queries.
         metadata : Optional Mapping of `str` to primitives
             This is an optional parameter of extra data to carry with the
             graph.  Entries in this mapping should be able to be serialized in
@@ -1265,9 +1296,63 @@ class _PipelineScaffolding:
             initInputs=taskInitInputs,
             initOutputs=taskInitOutputs,
             globalInitOutputs=globalInitOutputs,
+            registryDatasetTypes=self._get_registry_dataset_types(registry),
         )
         return graph
 
+    def _get_registry_dataset_types(self, registry: Registry) -> Iterable[DatasetType]:
+        """Make a list of all dataset types used by a graph as defined in
+        registry.
+        """
+        chain = [
+            self.initInputs,
+            self.initIntermediates,
+            self.initOutputs,
+            self.inputs,
+            self.intermediates,
+            self.outputs,
+            self.prerequisites,
+        ]
+        if self.globalInitOutputs is not None:
+            chain.append(self.globalInitOutputs)
+
+        # Collect names of all dataset types.
+        all_names: set[str] = set(dstype.name for dstype in itertools.chain(*chain))
+        dataset_types = {ds.name: ds for ds in registry.queryDatasetTypes(all_names)}
+
+        # Check for types that do not exist in registry yet:
+        # - inputs must exist
+        # - intermediates and outputs may not exist, but there must not be
+        #   more than one definition (e.g. differing in storage class)
+        # - prerequisites may not exist, treat it the same as outputs here
+        for dstype in itertools.chain(self.initInputs, self.inputs):
+            if dstype.name not in dataset_types:
+                raise MissingDatasetTypeError(f"Registry is missing an input dataset type {dstype}")
+
+        new_outputs: dict[str, set[DatasetType]] = defaultdict(set)
+        chain = [
+            self.initIntermediates,
+            self.initOutputs,
+            self.intermediates,
+            self.outputs,
+            self.prerequisites,
+        ]
+        if self.globalInitOutputs is not None:
+            chain.append(self.globalInitOutputs)
+        for dstype in itertools.chain(*chain):
+            if dstype.name not in dataset_types:
+                new_outputs[dstype.name].add(dstype)
+        for name, dstypes in new_outputs.items():
+            if len(dstypes) > 1:
+                raise ValueError(
+                    "Pipeline contains multiple definitions for a dataset type "
+                    f"which is not defined in registry yet: {dstypes}"
+                )
+            elif len(dstypes) == 1:
+                dataset_types[name] = dstypes.pop()
+
+        return dataset_types.values()
+
 
 # ------------------------
 #  Exported definitions --
@@ -1413,4 +1498,6 @@ class GraphBuilder:
                 constrainedByAllDatasets=condition,
                 resolveRefs=resolveRefs,
             )
-        return scaffolding.makeQuantumGraph(metadata=metadata, datastore=self.datastore)
+        return scaffolding.makeQuantumGraph(
+            registry=self.registry, metadata=metadata, datastore=self.datastore
+        )
