@@ -30,7 +30,6 @@ import logging
 import os
 import re
 import urllib.parse
-import warnings
 
 # -------------------------------
 #  Imports of standard modules --
@@ -67,6 +66,7 @@ from ._task_metadata import TaskMetadata
 from .config import PipelineTaskConfig
 from .configOverrides import ConfigOverrides
 from .connections import iterConnections
+from .connectionTypes import Input
 from .pipelineTask import PipelineTask
 from .task import _TASK_METADATA_TYPE
 
@@ -407,11 +407,9 @@ class Pipeline:
             # This is to support legacy pipelines during transition
             uri, num_replace = re.subn("[:](?!\\/\\/)", "#", uri)
             if num_replace:
-                warnings.warn(
-                    f"The pipeline file {uri} seems to use the legacy : to separate "
-                    "labels, this is deprecated and will be removed after June 2021, please use "
-                    "# instead.",
-                    category=FutureWarning,
+                raise ValueError(
+                    f"The pipeline file {uri} seems to use the legacy :"
+                    " to separate labels, please use # instead."
                 )
             if uri.count("#") > 1:
                 raise ValueError("Only one set of labels is allowed when specifying a pipeline to load")
@@ -859,6 +857,12 @@ class TaskDatasetTypes:
     at the Pipeline level.
     """
 
+    queryConstraints: NamedValueSet[DatasetType]
+    """Regular inputs that should not be used as constraints on the initial
+    QuantumGraph generation data ID query, according to their tasks
+    (`NamedValueSet`).
+    """
+
     prerequisites: NamedValueSet[DatasetType]
     """Dataset types that are prerequisite inputs to this Task.
 
@@ -1097,10 +1101,18 @@ class TaskDatasetTypes:
 
         outputs.freeze()
 
+        inputs = makeDatasetTypesSet("inputs", is_input=True)
+        queryConstraints = NamedValueSet(
+            inputs[c.name]
+            for c in cast(Iterable[Input], iterConnections(taskDef.connections, "inputs"))
+            if not c.deferGraphConstraint
+        )
+
         return cls(
             initInputs=makeDatasetTypesSet("initInputs", is_input=True),
             initOutputs=initOutputs,
-            inputs=makeDatasetTypesSet("inputs", is_input=True),
+            inputs=inputs,
+            queryConstraints=queryConstraints,
             prerequisites=makeDatasetTypesSet("prerequisiteInputs", is_input=True),
             outputs=outputs,
         )
@@ -1147,6 +1159,12 @@ class PipelineDatasetTypes:
     produced.
     """
 
+    queryConstraints: NamedValueSet[DatasetType]
+    """Regular inputs that should be used as constraints on the initial
+    QuantumGraph generation data ID query, according to their tasks
+    (`NamedValueSet`).
+    """
+
     prerequisites: NamedValueSet[DatasetType]
     """Dataset types that are prerequisite inputs for the full Pipeline.
 
@@ -1219,6 +1237,7 @@ class PipelineDatasetTypes:
         allInitInputs = NamedValueSet[DatasetType]()
         allInitOutputs = NamedValueSet[DatasetType]()
         prerequisites = NamedValueSet[DatasetType]()
+        queryConstraints = NamedValueSet[DatasetType]()
         byTask = dict()
         if include_packages:
             allInitOutputs.add(
@@ -1247,6 +1266,9 @@ class PipelineDatasetTypes:
             allInitInputs.update(thisTask.initInputs)
             allInitOutputs.update(thisTask.initOutputs)
             allInputs.update(thisTask.inputs)
+            # Inputs are query constraints if any task considers them a query
+            # constraint.
+            queryConstraints.update(thisTask.queryConstraints)
             prerequisites.update(thisTask.prerequisites)
             allOutputs.update(thisTask.outputs)
             byTask[taskDef.label] = thisTask
@@ -1299,11 +1321,14 @@ class PipelineDatasetTypes:
             s.freeze()
             return s
 
+        inputs = frozen(allInputs - allOutputs - intermediateComponents)
+
         return cls(
             initInputs=frozen(allInitInputs - allInitOutputs),
             initIntermediates=frozen(allInitInputs & allInitOutputs),
             initOutputs=frozen(allInitOutputs - allInitInputs),
-            inputs=frozen(allInputs - allOutputs - intermediateComponents),
+            inputs=inputs,
+            queryConstraints=frozen(queryConstraints & inputs),
             # If there are storage class differences in inputs and outputs
             # the intermediates have to choose priority. Here choose that
             # inputs to tasks much match the requested storage class by
